@@ -1,9 +1,9 @@
 #!/usr/bin/env python
-# $Id: unix_ftpd.py 977 2012-01-22 23:05:09Z g.rodola $
+# $Id: unix_ftpd.py 1174 2013-02-19 11:25:49Z g.rodola $
 
 #  pyftpdlib is released under the MIT license, reproduced below:
 #  ======================================================================
-#  Copyright (C) 2007-2012 Giampaolo Rodola' <g.rodola@gmail.com>
+#  Copyright (C) 2007-2013 Giampaolo Rodola' <g.rodola@gmail.com>
 #
 #                         All Rights Reserved
 #
@@ -36,19 +36,19 @@ It temporarily impersonate the system users every time they are going
 to perform a filesystem operations.
 """
 
-from pyftpdlib import ftpserver
-from pyftpdlib.contrib.authorizers import UnixAuthorizer
-from pyftpdlib.contrib.filesystems import UnixFilesystem
+from pyftpdlib.handlers import FTPHandler
+from pyftpdlib.servers import FTPServer
+from pyftpdlib.authorizers import UnixAuthorizer
+from pyftpdlib.filesystems import UnixFilesystem
 
 
 def main():
     authorizer = UnixAuthorizer(rejected_users=["root"], require_valid_shell=True)
-    ftp_handler = ftpserver.FTPHandler
-    ftp_handler.authorizer = authorizer
-    ftp_handler.abstracted_fs = UnixFilesystem
-    address = ('', 21)
-    ftpd = ftpserver.FTPServer(address, ftp_handler)
-    ftpd.serve_forever()
+    handler = FTPHandler
+    handler.authorizer = authorizer
+    handler.abstracted_fs = UnixFilesystem
+    server = FTPServer(('', 21), handler)
+    server.serve_forever()
 
 if __name__ == "__main__":
     main()
