@@ -1,9 +1,9 @@
 #!/usr/bin/env python
-# $Id: bench.py 986 2012-01-25 18:21:32Z g.rodola $
+# $Id: bench.py 1202 2013-04-04 16:23:09Z g.rodola $
 #
 #  pyftpdlib is released under the MIT license, reproduced below:
 #  ======================================================================
-#  Copyright (C) 2007-2012 Giampaolo Rodola' <g.rodola@gmail.com>
+#  Copyright (C) 2007-2013 Giampaolo Rodola' <g.rodola@gmail.com>
 #
 #                         All Rights Reserved
 #
@@ -48,45 +48,45 @@ Example usages:
 """
 
 # Some benchmarks (Linux 3.0.0, Intel core duo - 3.1 Ghz).
+
+# pyftpdlib 1.0.0:
 #
-# pyftpdlib 0.6.0:
+#   (starting with 6.7M of memory being used)
+#   STOR (client -> server)                              557.97 MB/sec  6.7M
+#   RETR (server -> client)                             1613.82 MB/sec  6.8M
+#   300 concurrent clients (connect, login)                1.20 secs    8.8M
+#   STOR (1 file with 300 idle clients)                  567.52 MB/sec  8.8M
+#   RETR (1 file with 300 idle clients)                 1561.41 MB/sec  8.8M
+#   300 concurrent clients (RETR 10.0M file)               3.26 secs    10.8M
+#   300 concurrent clients (STOR 10.0M file)               8.46 secs    12.6M
+#   300 concurrent clients (QUIT)                          0.07 secs
 #
-# (starting with 9M of RSS memory being used)
-# STOR (client -> server)                      528.02 MB/sec  9M RSS
-# RETR (server -> client)                      693.41 MB/sec  10M RSS
-# 200 concurrent clients (connect, login)        0.25 secs    10M RSS
-# 200 concurrent clients (RETR 10M file)         3.68 secs    11M RSS
-# 200 concurrent clients (STOR 10M file)         6.11 secs    11M RSS
-# 200 concurrent clients (quit)                  0.02 secs
 #
-# pyftpdlib 0.7.0:
-#  (starting with 6M of RSS memory being used)
-#   STOR (client -> server)                      512.80 MB/sec  9M RSS
-#   RETR (server -> client)                     1694.14 MB/sec  9M RSS
-#   200 concurrent clients (connect, login)        0.95 secs    11M RSS
-#   200 concurrent clients (RETR 10M file)         2.80 secs    12M RSS
-#   200 concurrent clients (STOR 10M file)         6.30 secs    13M RSS
-#   200 concurrent clients (quit)                  0.02 secs
+# proftpd 1.3.4a:
 #
-# proftpd 1.3.4rc2:
+#   (starting with 1.4M of memory being used)
+#   STOR (client -> server)                              554.67 MB/sec  3.2M
+#   RETR (server -> client)                             1517.12 MB/sec  3.2M
+#   300 concurrent clients (connect, login)                9.30 secs    568.6M
+#   STOR (1 file with 300 idle clients)                  484.11 MB/sec  570.6M
+#   RETR (1 file with 300 idle clients)                 1534.61 MB/sec  570.6M
+#   300 concurrent clients (RETR 10.0M file)               3.67 secs    568.6M
+#   300 concurrent clients (STOR 10.0M file)              11.21 secs    568.7M
+#   300 concurrent clients (QUIT)                          0.43 secs
 #
-#   (starting with 2M of RSS memory being used)
-#   STOR (client -> server)                      609.22 MB/sec  6M RSS
-#   RETR (server -> client)                     1313.77 MB/sec  6M RSS
-#   200 concurrent clients (connect, login)        7.53 secs    862M RSS
-#   200 concurrent clients (RETR 10M file)         2.76 secs    862M RSS
-#   200 concurrent clients (STOR 10M file)         6.39 secs    862M RSS
-#   200 concurrent clients (quit)                  0.23 secs
 #
 # vsftpd 2.3.2
 #
-#   (starting with 1M of RSS memory being used)
-#   STOR (client -> server)                      670.48 MB/sec  2M RSS
-#   RETR (server -> client)                     1505.18 MB/sec  2M RSS
-#   200 concurrent clients (connect, login)       12.61 secs    289M RSS
-#   200 concurrent clients (RETR 10M file)         2.30 secs    289M RSS
-#   200 concurrent clients (STOR 10M file)          N/A
-#   200 concurrent clients (quit)                  0.01
+#   (starting with 352.0K of memory being used)
+#   STOR (client -> server)                              607.23 MB/sec  816.0K
+#   RETR (server -> client)                             1506.59 MB/sec  816.0K
+#   300 concurrent clients (connect, login)               18.91 secs    140.9M
+#   STOR (1 file with 300 idle clients)                  618.99 MB/sec  141.4M
+#   RETR (1 file with 300 idle clients)                 1402.48 MB/sec  141.4M
+#   300 concurrent clients (RETR 10.0M file)               3.64 secs    140.9M
+#   300 concurrent clients (STOR 10.0M file)               9.74 secs    140.9M
+#   300 concurrent clients (QUIT)                          0.00 secs
+
 
 
 from __future__ import with_statement, division
@@ -118,6 +118,10 @@ BUFFER_LEN = 8192
 SERVER_PROC = None
 
 
+def print_(s):
+    sys.stdout.write(s + '\n')
+    sys.stdout.flush()
+
 # http://goo.gl/6V8Rm
 def hilite(string, ok=True, bold=False):
     """Return an highlighted version of 'string'."""
@@ -139,15 +143,15 @@ if not sys.stdout.isatty() or os.name != 'posix':
 server_memory = []
 
 def print_bench(what, value, unit=""):
-    s = "%s %s %-8s" % (hilite("%-42s" % what, ok=None, bold=0),
+    s = "%s %s %-8s" % (hilite("%-50s" % what, ok=None, bold=0),
                       hilite("%8.2f" % value),
                       unit)
     if server_memory:
-        s += "%s RSS" % hilite(server_memory.pop())
-    print s.strip()
+        s += "%s" % hilite(server_memory.pop())
+    print_(s.strip())
 
 # http://goo.gl/zeJZl
-def bytes2human(n, format="%(value)i%(symbol)s"):
+def bytes2human(n, format="%(value).1f%(symbol)s"):
     """
     >>> bytes2human(10000)
     '9K'
@@ -183,11 +187,27 @@ def human2bytes(s):
     return int(num * prefix[letter])
 
 def register_memory():
-    """Register RSS memory used by server process and its children."""
+    """Register an approximation of memory used by FTP server process
+    and all of its children.
+    """
+    # XXX How to get a reliable representation of memory being used is
+    # not clear. (rss - shared) seems kind of ok but we might also use
+    # the private working set via get_memory_maps().private*.
+    def get_mem(proc):
+        if os.name == 'posix':
+            mem = proc.get_ext_memory_info()
+            counter = mem.rss
+            if 'shared' in mem._fields:
+                counter -= mem.shared
+            return counter
+        else:
+            # TODO figure out what to do on Windows
+            return proc.get_memory_info().rss
+
     if SERVER_PROC is not None:
-        mem = SERVER_PROC.get_memory_info().rss
+        mem = get_mem(SERVER_PROC)
         for child in SERVER_PROC.get_children():
-            mem += child.get_memory_info().rss
+            mem += get_mem(child)
         server_memory.append(bytes2human(mem))
 
 def timethis(what):
@@ -410,15 +430,13 @@ def main():
                 raise ImportError("-p option requires psutil module")
             SERVER_PROC = psutil.Process(options.pid)
 
-    def bench_stor():
+    def bench_stor(title="STOR (client -> server)"):
         bytes = bytes_per_second(connect(), retr=False)
-        print_bench("STOR (client -> server)",
-                    round(bytes / 1024.0 / 1024.0, 2), "MB/sec")
+        print_bench(title, round(bytes / 1024.0 / 1024.0, 2), "MB/sec")
 
-    def bench_retr():
+    def bench_retr(title="RETR (server -> client)"):
         bytes = bytes_per_second(connect(), retr=True)
-        print_bench("RETR (server -> client)",
-                    round(bytes / 1024.0 / 1024.0, 2), "MB/sec")
+        print_bench(title, round(bytes / 1024.0 / 1024.0, 2), "MB/sec")
 
     def bench_multi():
         howmany = options.clients
@@ -466,10 +484,12 @@ def main():
         def bench_multi_quit(clients):
             for ftp in clients:
                 AsyncQuit(ftp.sock)
-            with timethis("%i concurrent clients (quit)" % howmany):
+            with timethis("%i concurrent clients (QUIT)" % howmany):
                 asyncore.loop(use_poll=True)
 
         clients = bench_multi_connect()
+        bench_stor("STOR (1 file with %s idle clients)" % len(clients))
+        bench_retr("RETR (1 file with %s idle clients)" % len(clients))
         bench_multi_retr(clients)
         bench_multi_stor(clients)
         bench_multi_quit(clients)
@@ -486,8 +506,8 @@ def main():
     # start benchmark
     if SERVER_PROC is not None:
         register_memory()
-        print "(starting with %s of RSS memory being used)" \
-               % hilite(server_memory.pop())
+        print_("(starting with %s of memory being used)" \
+               % hilite(server_memory.pop()))
     if options.benchmark == 'transfer':
         bench_stor()
         bench_retr()
