@@ -1,9 +1,9 @@
 #!/usr/bin/env python
-# $Id: test_contrib.py 977 2012-01-22 23:05:09Z g.rodola $
+# $Id: test_contrib.py 1219 2013-04-19 14:35:41Z g.rodola $
 
 #  pyftpdlib is released under the MIT license, reproduced below:
-#  ======================================================================
-#  Copyright (C) 2007-2012 Giampaolo Rodola' <g.rodola@gmail.com>
+#  ========================================================================
+#  Copyright (C) 2007-2013 Giampaolo Rodola' <g.rodola@gmail.com>
 #
 #                         All Rights Reserved
 #
@@ -28,10 +28,11 @@
 # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 # OTHER DEALINGS IN THE SOFTWARE.
 #
-#  ======================================================================
+#  ========================================================================
 
-"""Tests for pyftpdlib.contrib namespace: handlers.py, authorizers.py
-and filesystems.py modules.
+"""
+Tests for pyftpdlib.contrib namespace: handlers.py, authorizers.py,
+filesystems.py and servers.py modules.
 """
 
 import ftplib
@@ -56,15 +57,41 @@ try:
 except ImportError:
     pass
 
-from pyftpdlib import ftpserver
-from pyftpdlib.contrib import authorizers
-from pyftpdlib.contrib import handlers
-from pyftpdlib.contrib import filesystems
+from pyftpdlib.authorizers import AuthenticationFailed, AuthorizerError
+from pyftpdlib import authorizers
+from pyftpdlib import handlers
+from pyftpdlib import filesystems
+from pyftpdlib import servers
+from pyftpdlib._compat import b, getcwdu, unicode
 from test_ftpd import *
 
 
 FTPS_SUPPORT = hasattr(ftplib, 'FTP_TLS') and hasattr(handlers, 'TLS_FTPHandler')
 CERTFILE = os.path.abspath(os.path.join(os.path.dirname(__file__), 'keycert.pem'))
+MPROCESS_SUPPORT = hasattr(servers, 'MultiprocessFTPServer')
+
+
+# =====================================================================
+# --- Mixin tests
+# =====================================================================
+
+# What we're going to do here is repeat the original functional tests
+# defined in test_ftpd.py but by using different FTP server
+# configurations.
+#
+# In case of FTPS we secure both control and data connections before
+# running any test.
+# Same story for ThreadedFTPServer which will be used instead of
+# base FTPServer class.
+#
+# This is useful as we reuse the existent functional tests which are
+# supposed to work no matter if the underlying protocol is FTP or FTPS,
+# or if the concurrency module used is asynchronous or based on
+# multiple threads or processes (fork).
+
+# =====================================================================
+# --- FTPS mixin tests
+# =====================================================================
 
 if FTPS_SUPPORT:
     class FTPSClient(ftplib.FTP_TLS):
@@ -87,20 +114,6 @@ else:
     class TLSTestMixin:
         pass
 
-# --- FTPS mixin tests
-#
-# What we're going to do here is repeating the original tests
-# defined in test_ftpd.py but securing both control and data
-# connections first.
-# The tests are exactly the same, the only difference is that
-# different classes are used (TLS_FPTHandler for the server
-# and ftplib.FTP_TLS for the client) and everything goes through
-# FTPS instead of clear-text FTP.
-# This is very useful as we entirely reuse the existent test code
-# base which is very large (more than 100 tests) and covers a lot
-# of cases which are supposed to work no matter if the protocol
-# is FTP or FTPS.
-
 class TestFtpAuthenticationTLSMixin(TLSTestMixin, TestFtpAuthentication): pass
 class TestTFtpDummyCmdsTLSMixin(TLSTestMixin, TestFtpDummyCmds): pass
 class TestFtpCmdsSemanticTLSMixin(TLSTestMixin, TestFtpCmdsSemantic): pass
@@ -108,6 +121,7 @@ class TestFtpFsOperationsTLSMixin(TLSTestMixin, TestFtpFsOperations): pass
 class TestFtpStoreDataTLSMixin(TLSTestMixin, TestFtpStoreData): pass
 class TestFtpRetrieveDataTLSMixin(TLSTestMixin, TestFtpRetrieveData): pass
 class TestFtpListingCmdsTLSMixin(TLSTestMixin, TestFtpListingCmds): pass
+
 class TestFtpAbortTLSMixin(TLSTestMixin, TestFtpAbort):
     def test_oob_abor(self): pass
 
@@ -120,18 +134,87 @@ class TestCallbacksTLSMixin(TLSTestMixin, TestCallbacks):
     def test_on_file_sent(self): pass
     def test_on_incomplete_file_received(self): pass
     def test_on_incomplete_file_sent(self): pass
+    def test_on_connect(self): pass
+    def test_on_disconnect(self): pass
     def test_on_login(self): pass
     def test_on_login_failed(self): pass
     def test_on_logout_quit(self): pass
     def test_on_logout_rein(self): pass
     def test_on_logout_user_issued_twice(self): pass
 
-
 class TestIPv4EnvironmentTLSMixin(TLSTestMixin, TestIPv4Environment): pass
 class TestIPv6EnvironmentTLSMixin(TLSTestMixin, TestIPv6Environment): pass
 class TestCornerCasesTLSMixin(TLSTestMixin, TestCornerCases): pass
 
 
+# =====================================================================
+# --- threaded FTP server mixin tests
+# =====================================================================
+
+class TFTPd(FTPd):
+    server_class = servers.ThreadedFTPServer
+
+class ThreadFTPTestMixin:
+    server_class = TFTPd
+
+class TestFtpAuthenticationThreadMixin(ThreadFTPTestMixin, TestFtpAuthentication): pass
+class TestTFtpDummyCmdsThreadMixin(ThreadFTPTestMixin, TestFtpDummyCmds): pass
+class TestFtpCmdsSemanticThreadMixin(ThreadFTPTestMixin, TestFtpCmdsSemantic): pass
+class TestFtpFsOperationsThreadMixin(ThreadFTPTestMixin, TestFtpFsOperations): pass
+class TestFtpStoreDataThreadMixin(ThreadFTPTestMixin, TestFtpStoreData): pass
+class TestFtpRetrieveDataThreadMixin(ThreadFTPTestMixin, TestFtpRetrieveData): pass
+class TestFtpListingCmdsThreadMixin(ThreadFTPTestMixin, TestFtpListingCmds): pass
+class TestFtpAbortThreadMixin(ThreadFTPTestMixin, TestFtpAbort): pass
+#class TestTimeoutsThreadMixin(ThreadFTPTestMixin, TestTimeouts):
+#    def test_data_timeout_not_reached(self): pass
+#class TestConfigurableOptionsThreadMixin(ThreadFTPTestMixin, TestConfigurableOptions): pass
+class TestCallbacksThreadMixin(ThreadFTPTestMixin, TestCallbacks): pass
+class TestIPv4EnvironmentThreadMixin(ThreadFTPTestMixin, TestIPv4Environment): pass
+class TestIPv6EnvironmentThreadMixin(ThreadFTPTestMixin, TestIPv6Environment): pass
+class TestCornerCasesThreadMixin(ThreadFTPTestMixin, TestCornerCases): pass
+class TestFTPServerThreadMixin(ThreadFTPTestMixin, TestFTPServer): pass
+
+
+# =====================================================================
+# --- multiprocess FTP server mixin tests
+# =====================================================================
+
+if MPROCESS_SUPPORT:
+    class MultiProcFTPd(FTPd):
+        server_class = servers.MultiprocessFTPServer
+
+    class MProcFTPTestMixin:
+        server_class = MultiProcFTPd
+else:
+    class MProcFTPTestMixin:
+        pass
+
+class TestFtpAuthenticationMProcMixin(MProcFTPTestMixin, TestFtpAuthentication): pass
+class TestTFtpDummyCmdsMProcMixin(MProcFTPTestMixin, TestFtpDummyCmds): pass
+class TestFtpCmdsSemanticMProcMixin(MProcFTPTestMixin, TestFtpCmdsSemantic): pass
+
+class TestFtpFsOperationsMProcMixin(MProcFTPTestMixin, TestFtpFsOperations):
+    def test_unforeseen_mdtm_event(self):
+        pass
+
+class TestFtpStoreDataMProcMixin(MProcFTPTestMixin, TestFtpStoreData): pass
+class TestFtpRetrieveDataMProcMixin(MProcFTPTestMixin, TestFtpRetrieveData): pass
+class TestFtpListingCmdsMProcMixin(MProcFTPTestMixin, TestFtpListingCmds): pass
+class TestFtpAbortMProcMixin(MProcFTPTestMixin, TestFtpAbort): pass
+#class TestTimeoutsMProcMixin(MProcFTPTestMixin, TestTimeouts):
+#    def test_data_timeout_not_reached(self): pass
+#class TestConfigurableOptionsMProcMixin(MProcFTPTestMixin, TestConfigurableOptions): pass
+#class TestCallbacksMProcMixin(MProcFTPTestMixin, TestCallbacks): pass
+class TestIPv4EnvironmentMProcMixin(MProcFTPTestMixin, TestIPv4Environment): pass
+class TestIPv6EnvironmentMProcMixin(MProcFTPTestMixin, TestIPv6Environment): pass
+class TestCornerCasesMProcMixin(MProcFTPTestMixin, TestCornerCases): pass
+class TestFTPServerMProcMixin(MProcFTPTestMixin, TestFTPServer): pass
+
+
+# =====================================================================
+# dedicated FTPs tests
+# =====================================================================
+
 class TestFTPS(unittest.TestCase):
     """Specific tests fot TSL_FTPHandler class."""
 
@@ -153,7 +236,8 @@ class TestFTPS(unittest.TestCase):
     def assertRaisesWithMsg(self, excClass, msg, callableObj, *args, **kwargs):
         try:
             callableObj(*args, **kwargs)
-        except excClass, why:
+        except excClass:
+            why = sys.exc_info()[1]
             if str(why) == msg:
                 return
             raise self.failureException("%s != %s" % (str(why), msg))
@@ -162,7 +246,7 @@ class TestFTPS(unittest.TestCase):
                 excName = excClass.__name__
             else:
                 excName = str(excClass)
-            raise self.failureException, "%s not raised" % excName
+            raise self.failureException("%s not raised" % excName)
 
     def test_auth(self):
         # unsecured
@@ -196,19 +280,27 @@ class TestFTPS(unittest.TestCase):
         # secured
         self.client.prot_p()
         sock = self.client.transfercmd('list')
-        while 1:
-            if not sock.recv(1024):
-                self.client.voidresp()
-                break
-        self.assertTrue(isinstance(sock, ssl.SSLSocket))
-        # unsecured
-        self.client.prot_c()
+        try:
+            sock.settimeout(TIMEOUT)
+            while 1:
+                if not sock.recv(1024):
+                    self.client.voidresp()
+                    break
+            self.assertTrue(isinstance(sock, ssl.SSLSocket))
+            # unsecured
+            self.client.prot_c()
+        finally:
+            sock.close()
         sock = self.client.transfercmd('list')
-        while 1:
-            if not sock.recv(1024):
-                self.client.voidresp()
-                break
-        self.assertFalse(isinstance(sock, ssl.SSLSocket))
+        try:
+            sock.settimeout(TIMEOUT)
+            while 1:
+                if not sock.recv(1024):
+                    self.client.voidresp()
+                    break
+            self.assertFalse(isinstance(sock, ssl.SSLSocket))
+        finally:
+            sock.close()
 
     def test_feat(self):
         feat = self.client.sendcmd('feat')
@@ -220,17 +312,19 @@ class TestFTPS(unittest.TestCase):
         self.client.login()
         try:
             sock = self.client.sock.unwrap()
-        except socket.error, err:
+        except socket.error:
+            err = sys.exc_info()[1]
             if err.errno == 0:
                 return
             raise
-        sock.sendall('noop')
+        sock.settimeout(TIMEOUT)
+        sock.sendall(b('noop'))
         try:
             chunk = sock.recv(1024)
         except socket.error:
             pass
         else:
-            self.assertEqual(chunk, "")
+            self.assertEqual(chunk, b(""))
 
     def test_tls_control_required(self):
         self.server.handler.tls_control_required = True
@@ -284,7 +378,9 @@ class TestFTPS(unittest.TestCase):
             self.client.ssl_version = ssl.PROTOCOL_SSLv2
 
 
-# --- System dependant authorizers tests
+# =====================================================================
+# --- authorizer
+# =====================================================================
 
 class SharedAuthorizerTests(unittest.TestCase):
     """Tests valid for both UnixAuthorizer and WindowsAuthorizer for
@@ -321,25 +417,27 @@ class SharedAuthorizerTests(unittest.TestCase):
     def assertRaisesWithMsg(self, excClass, msg, callableObj, *args, **kwargs):
         try:
             callableObj(*args, **kwargs)
-        except excClass, why:
+        except excClass:
+            why = sys.exc_info()[1]
             if str(why) == msg:
                 return
             raise self.failureException("%s != %s" % (str(why), msg))
         else:
             if hasattr(excClass,'__name__'): excName = excClass.__name__
             else: excName = str(excClass)
-            raise self.failureException, "%s not raised" % excName
+            raise self.failureException("%s not raised" % excName)
 
     # --- /utils
 
     def test_get_home_dir(self):
         auth = self.authorizer_class()
         home = auth.get_home_dir(self.get_current_user())
+        self.assertTrue(isinstance(home, unicode))
         nonexistent_user = self.get_nonexistent_user()
         self.assertTrue(os.path.isdir(home))
         if auth.has_user('nobody'):
             home = auth.get_home_dir('nobody')
-        self.assertRaises(ftpserver.AuthorizerError,
+        self.assertRaises(AuthorizerError,
                           auth.get_home_dir, nonexistent_user)
 
     def test_has_user(self):
@@ -360,8 +458,10 @@ class SharedAuthorizerTests(unittest.TestCase):
             auth = self.authorizer_class()
         current_user = self.get_current_user()
         nonexistent_user = self.get_nonexistent_user()
-        self.assertFalse(auth.validate_authentication(current_user, 'wrongpasswd'))
-        self.assertFalse(auth.validate_authentication(nonexistent_user, 'bar'))
+        self.assertRaises(AuthenticationFailed,
+              auth.validate_authentication, current_user, 'wrongpasswd', None)
+        self.assertRaises(AuthenticationFailed,
+                  auth.validate_authentication, nonexistent_user, 'bar', None)
 
     def test_impersonate_user(self):
         auth = self.authorizer_class()
@@ -369,7 +469,7 @@ class SharedAuthorizerTests(unittest.TestCase):
         try:
             if self.authorizer_class.__name__ == 'UnixAuthorizer':
                 auth.impersonate_user(self.get_current_user(), '')
-                self.assertRaises(ftpserver.AuthorizerError,
+                self.assertRaises(AuthorizerError,
                                   auth.impersonate_user, nonexistent_user, 'pwd')
             else:
                 self.assertRaises(Win32ExtError,
@@ -401,27 +501,28 @@ class SharedAuthorizerTests(unittest.TestCase):
 
     def test_error_options(self):
         wrong_user = self.get_nonexistent_user()
-        self.assertRaisesWithMsg(ValueError,
+        self.assertRaisesWithMsg(AuthorizerError,
            "rejected_users and allowed_users options are mutually exclusive",
            self.authorizer_class, allowed_users=['foo'], rejected_users=['bar'])
-        self.assertRaisesWithMsg(ValueError,
+        self.assertRaisesWithMsg(AuthorizerError,
                              'invalid username "anonymous"',
                              self.authorizer_class, allowed_users=['anonymous'])
-        self.assertRaisesWithMsg(ValueError,
+        self.assertRaisesWithMsg(AuthorizerError,
                             'invalid username "anonymous"',
                             self.authorizer_class, rejected_users=['anonymous'])
-        self.assertRaisesWithMsg(ValueError,
+        self.assertRaisesWithMsg(AuthorizerError,
                             'unknown user %s' % wrong_user,
                             self.authorizer_class, allowed_users=[wrong_user])
-        self.assertRaisesWithMsg(ValueError, 'unknown user %s' % wrong_user,
+        self.assertRaisesWithMsg(AuthorizerError, 'unknown user %s' % wrong_user,
                             self.authorizer_class, rejected_users=[wrong_user])
 
     def test_override_user_password(self):
         auth = self.authorizer_class()
         user = self.get_current_user()
         auth.override_user(user, password='foo')
-        self.assertTrue(auth.validate_authentication(user, 'foo'))
-        self.assertFalse(auth.validate_authentication(user, 'bar'))
+        auth.validate_authentication(user, 'foo', None)
+        self.assertRaises(AuthenticationFailed(auth.validate_authentication,
+                                               user, 'bar', None))
         # make sure other settings keep using default values
         self.assertEqual(auth.get_home_dir(user), self.get_current_user_homedir())
         self.assertEqual(auth.get_perms(user), "elradfmw")
@@ -431,7 +532,7 @@ class SharedAuthorizerTests(unittest.TestCase):
     def test_override_user_homedir(self):
         auth = self.authorizer_class()
         user = self.get_current_user()
-        dir = os.path.dirname(os.getcwd())
+        dir = os.path.dirname(getcwdu())
         auth.override_user(user, homedir=dir)
         self.assertEqual(auth.get_home_dir(user), dir)
         # make sure other settings keep using default values
@@ -474,10 +575,10 @@ class SharedAuthorizerTests(unittest.TestCase):
                 another_user = x
                 break
         nonexistent_user = self.get_nonexistent_user()
-        self.assertRaisesWithMsg(ValueError,
+        self.assertRaisesWithMsg(AuthorizerError,
                                 "at least one keyword argument must be specified",
                                 auth.override_user, this_user)
-        self.assertRaisesWithMsg(ValueError,
+        self.assertRaisesWithMsg(AuthorizerError,
                                  'no such user %s' % nonexistent_user,
                                  auth.override_user, nonexistent_user, perm='r')
         if self.authorizer_class.__name__ == 'UnixAuthorizer':
@@ -486,7 +587,7 @@ class SharedAuthorizerTests(unittest.TestCase):
         else:
             auth = self.authorizer_class(allowed_users=[this_user])
         auth.override_user(this_user, perm='r')
-        self.assertRaisesWithMsg(ValueError,
+        self.assertRaisesWithMsg(AuthorizerError,
                                  '%s is not an allowed user' % another_user,
                                  auth.override_user, another_user, perm='r')
         if self.authorizer_class.__name__ == 'UnixAuthorizer':
@@ -495,14 +596,18 @@ class SharedAuthorizerTests(unittest.TestCase):
         else:
             auth = self.authorizer_class(rejected_users=[this_user])
         auth.override_user(another_user, perm='r')
-        self.assertRaisesWithMsg(ValueError,
+        self.assertRaisesWithMsg(AuthorizerError,
                                  '%s is not an allowed user' % this_user,
                                  auth.override_user, this_user, perm='r')
-        self.assertRaisesWithMsg(ValueError,
+        self.assertRaisesWithMsg(AuthorizerError,
                                  "can't assign password to anonymous user",
                                  auth.override_user, "anonymous", password='foo')
 
 
+# =====================================================================
+# --- UNIX authorizer
+# =====================================================================
+
 class TestUnixAuthorizer(SharedAuthorizerTests):
     """Unix authorizer specific tests."""
 
@@ -533,19 +638,21 @@ class TestUnixAuthorizer(SharedAuthorizerTests):
     def test_validate_authentication(self):
         # we can only test for invalid credentials
         auth = authorizers.UnixAuthorizer(require_valid_shell=False)
-        ret = auth.validate_authentication('?!foo', '?!foo')
-        self.assertFalse(ret)
+        self.assertRaises(AuthenticationFailed,
+                          auth.validate_authentication, '?!foo', '?!foo', None)
         auth = authorizers.UnixAuthorizer(require_valid_shell=True)
-        ret = auth.validate_authentication('?!foo', '?!foo')
-        self.assertFalse(ret)
+        self.assertRaises(AuthenticationFailed,
+                          auth.validate_authentication, '?!foo', '?!foo', None)
 
     def test_validate_authentication_anonymous(self):
         current_user = self.get_current_user()
         auth = authorizers.UnixAuthorizer(anonymous_user=current_user,
                                           require_valid_shell=False)
-        self.assertFalse(auth.validate_authentication('foo', 'passwd'))
-        self.assertFalse(auth.validate_authentication(current_user, 'passwd'))
-        self.assertTrue(auth.validate_authentication('anonymous', 'passwd'))
+        self.assertRaises(AuthenticationFailed,
+                          auth.validate_authentication, 'foo', 'passwd', None)
+        self.assertRaises(AuthenticationFailed,
+                     auth.validate_authentication, current_user, 'passwd', None)
+        auth.validate_authentication('anonymous', 'passwd', None)
 
     def test_require_valid_shell(self):
 
@@ -560,7 +667,7 @@ class TestUnixAuthorizer(SharedAuthorizerTests):
             self.fail("no user found")
 
         user = get_fake_shell_user()
-        self.assertRaisesWithMsg(ValueError,
+        self.assertRaisesWithMsg(AuthorizerError,
                              "user %s has not a valid shell" % user,
                              authorizers.UnixAuthorizer, allowed_users=[user])
         # commented as it first fails for invalid home
@@ -570,8 +677,8 @@ class TestUnixAuthorizer(SharedAuthorizerTests):
         auth = authorizers.UnixAuthorizer()
         self.assertTrue(auth._has_valid_shell(self.get_current_user()))
         self.assertFalse(auth._has_valid_shell(user))
-        self.assertRaisesWithMsg(ValueError,
-                                 "user %s has not a valid shell" % user,
+        self.assertRaisesWithMsg(AuthorizerError,
+                                 "User %s doesn't have a valid shell." % user,
                                  auth.override_user, user, perm='r')
 
     def test_not_root(self):
@@ -579,13 +686,17 @@ class TestUnixAuthorizer(SharedAuthorizerTests):
         auth = self.authorizer_class()
         try:
             auth.impersonate_user('nobody', '')
-            self.assertRaisesWithMsg(ftpserver.AuthorizerError,
+            self.assertRaisesWithMsg(AuthorizerError,
                                      "super user privileges are required",
                                      authorizers.UnixAuthorizer)
         finally:
             auth.terminate_impersonation('nobody')
 
 
+# =====================================================================
+# --- Windows authorizer
+# =====================================================================
+
 class TestWindowsAuthorizer(SharedAuthorizerTests):
     """Windows authorizer specific tests."""
 
@@ -597,23 +708,30 @@ class TestWindowsAuthorizer(SharedAuthorizerTests):
                        anonymous_user=user, anonymous_password='$|1wrongpasswd')
 
 
+# =====================================================================
+# --- UNIX filesystem
+# =====================================================================
+
 if os.name == 'posix':
     class TestUnixFilesystem(unittest.TestCase):
 
         def test_case(self):
-            root = os.getcwd()
+            root = getcwdu()
             fs = filesystems.UnixFilesystem(root, None)
             self.assertEqual(fs.root, root)
             self.assertEqual(fs.cwd, root)
             cdup = os.path.dirname(root)
-            self.assertEqual(fs.ftp2fs('..'), cdup)
+            self.assertEqual(fs.ftp2fs(u('..')), cdup)
             self.assertEqual(fs.fs2ftp(root), root)
 
 
+# =====================================================================
+# --- main
+# =====================================================================
+
 def test_main():
     test_suite = unittest.TestSuite()
     tests = []
-    warns = []
 
     # FTPS tests
     if FTPS_SUPPORT:
@@ -638,32 +756,72 @@ def test_main():
         tests += ftps_tests
     else:
         if sys.version_info < (2, 7):
-            warns.append("FTPS tests skipped (requires python 2.7)")
+            warn("FTPS tests skipped (requires python 2.7)")
         elif ssl is None:
-            warns.append("FTPS tests skipped (requires ssl module)")
+            warn("FTPS tests skipped (requires ssl module)")
         elif not hasattr(handlers, 'TLS_FTPHandler'):
-            warns.append("FTPS tests skipped (requires PyOpenSSL module)")
+            warn("FTPS tests skipped (requires PyOpenSSL module)")
         else:
-            warns.append("FTPS tests skipped")
+            warn("FTPS tests skipped")
 
-    # authorizers tests
+    # threaded FTP server tests
+    ftp_thread_tests = [
+        TestFtpAuthenticationThreadMixin,
+        TestTFtpDummyCmdsThreadMixin,
+        TestFtpCmdsSemanticThreadMixin,
+        TestFtpFsOperationsThreadMixin,
+        TestFtpStoreDataThreadMixin,
+        TestFtpRetrieveDataThreadMixin,
+        TestFtpListingCmdsThreadMixin,
+        TestFtpAbortThreadMixin,
+        #TestTimeoutsThreadMixin,
+        #TestConfigurableOptionsThreadMixin,
+        TestCallbacksThreadMixin,
+        TestCornerCasesThreadMixin,
+        TestFTPServerThreadMixin,
+    ]
+    tests += ftp_thread_tests
+
+    # multi process FTP server tests
+    if MPROCESS_SUPPORT:
+        ftp_mproc_tests = [
+            TestFtpAuthenticationMProcMixin,
+            TestTFtpDummyCmdsMProcMixin,
+            TestFtpCmdsSemanticMProcMixin,
+            TestFtpFsOperationsMProcMixin,
+            TestFtpStoreDataMProcMixin,
+            TestFtpRetrieveDataMProcMixin,
+            TestFtpListingCmdsMProcMixin,
+            TestFtpAbortMProcMixin,
+            #TestTimeoutsMProcMixin,
+            #TestConfigurableOptionsMProcMixin,
+            #TestCallbacksMProcMixin,
+            TestCornerCasesMProcMixin,
+            TestFTPServerMProcMixin,
+        ]
+        tests += ftp_mproc_tests
+
+    # POSIX tests
     if os.name == 'posix':
+        tests.append(TestUnixFilesystem)
+
         if hasattr(authorizers, "UnixAuthorizer"):
             try:
                 authorizers.UnixAuthorizer()
-            except ftpserver.AuthorizerError:  # not root
-                warns.append("UnixAuthorizer tests skipped (root privileges are "
-                             "required)")
+            except AuthorizerError:  # not root
+                warn("UnixAuthorizer tests skipped (root privileges are " \
+                     "required)")
             else:
                 tests.append(TestUnixAuthorizer)
         else:
             try:
                 import spwd
             except ImportError:
-                warns.append("UnixAuthorizer tests skipped (spwd module is "
-                             "missing")
+                warn("UnixAuthorizer tests skipped (spwd module is missing")
             else:
-                warns.append("UnixAuthorizer tests skipped")
+                warn("UnixAuthorizer tests skipped")
+
+    # Windows tests
     elif os.name in ('nt', 'ce'):
         if hasattr(authorizers, "WindowsAuthorizer"):
             tests.append(TestWindowsAuthorizer)
@@ -671,25 +829,18 @@ def test_main():
             try:
                 import win32api
             except ImportError:
-                warns.append("WindowsAuthorizer tests skipped (pywin32 extension "
-                             "is required)")
+                warn("WindowsAuthorizer tests skipped (pywin32 extension " \
+                     "is required)")
             else:
-                warns.append("WindowsAuthorizer tests skipped")
-
-    if os.name == 'posix':
-        tests.append(TestUnixFilesystem)
+                warn("WindowsAuthorizer tests skipped")
 
     for test in tests:
         test_suite.addTest(unittest.makeSuite(test))
     try:
-        unittest.TextTestRunner(verbosity=2).run(test_suite)
-    except:
-        # in case of KeyboardInterrupt grant that the threaded FTP
-        # server running in background gets stopped
-        asyncore.socket_map.clear()
-        raise
-    for warn in warns:
-        warnings.warn(warn, RuntimeWarning)
+        result = unittest.TextTestRunner(verbosity=2).run(test_suite)
+    finally:
+        cleanup()
+    return result
 
 if __name__ == '__main__':
-    test_main()
+    sys.exit(not test_main().wasSuccessful())
